#include "ElementStorage.h"

#include <arviz_godot/lib/avi/connection/LiveStorageAdapter.h>
#include <arviz_godot/lib/avi/connection/ReplayStorageAdapter.h>

avi::connection::ElementStorage::ElementStorage(avi::connection::Component& component) :
    component(component), storage(std::make_unique<LiveStorageAdapter>(component))
{
}

armarx::viz::data::LayerUpdates
avi::connection::ElementStorage::getNext()
{
    if (switchToLive)
    {
        while(!storage->finishedProcessing()) {}
        storage = std::make_unique<LiveStorageAdapter>(component);
        switchToLive = false;
    }
    else if (switchToReplay)
    {
        storage = std::make_unique<ReplayStorageAdapter>(component, clock, replay);
        storage->setup();
        switchToReplay = false;
    }

    std::lock_guard<std::mutex> lock(feedbackMutex);

    storage->setInteraction(feedbackVec);
    feedbackVec = {};

    return storage->getLayerUpdates();
}

bool
avi::connection::ElementStorage::isConnected() const
{
    return component.isConnected();
}

bool
avi::connection::ElementStorage::isLive() const
{
    return live;
}

bool
avi::connection::ElementStorage::isRecording() const
{
    return recording;
}

std::string
avi::connection::ElementStorage::recordReplay(const std::string& name)
{
    if (component.isConnected())
    {
        std::string returnString = component.getStorageInterface()->startRecording(name);
        recording = true;

        if(!returnString.empty()) {
            return returnString;
        }

        return std::string();
    }

    return "NOT_CONNECTED";
}

void
avi::connection::ElementStorage::endRecording()
{
    recording = false;

    if (component.isConnected())
    {
        component.getStorageInterface()->stopRecording();
    }
}

std::vector<std::string>
avi::connection::ElementStorage::getAllReplays()
{
    return storage->getAllReplays();
}

void
avi::connection::ElementStorage::useLiveTiming()
{
    live = true;
    switchToLive = true;
}

void
avi::connection::ElementStorage::useReplayTiming(const std::string& replay,
                                                 std::shared_ptr<avi::connection::Clock> clock)
{
    live = false;
    switchToReplay = true;

    this->replay = replay;
    this->clock = clock;
}

void
avi::connection::ElementStorage::addInteraction(avi::connection::Interaction interaction)
{
    std::lock_guard<std::mutex> lock(feedbackMutex);
    feedbackVec.push_back(interaction);
}

avi::connection::StorageModeSwitch::recordingInfo
avi::connection::ElementStorage::getRecordingInformation(const std::string name)
{
    StorageModeSwitch::recordingInfo recordingInfo;
    armarx::viz::data::RecordingsInfo armarxRecordingsInfo;

    std::vector<armarx::viz::data::Recording> replays;

    if (component.isConnected())
    {
        armarxRecordingsInfo = component.getStorageInterface()->getAllRecordings();
        replays = armarxRecordingsInfo.recordings;
        recordingPath = armarxRecordingsInfo.recordingsPath + "/" + name;
    }

    for (const auto& replay : replays)
    {
        if (replay.id == name)
        {
            recordingInfo.recordingID = name;
            recordingInfo.revisionStart = replay.firstRevision;
            recordingInfo.revisionEnd = replay.lastRevision;
            recordingInfo.batches = replay.batchHeaders.size();
            recordingInfo.firstTimeStamp = replay.firstTimestampInMicroSeconds;
            recordingInfo.lastTimeStamp = replay.lastTimestampInMicroSeconds;
        }
    }

    return recordingInfo;
}
std::filesystem::path
avi::connection::ElementStorage::getRecordingPath()
{
    return recordingPath;
}
