#include "ReplayStorageAdapter.h"

#include <utility>
#include <OS.hpp>

avi::connection::ReplayStorageAdapter::ReplayStorageAdapter(
    avi::connection::Component& component,
    std::shared_ptr<avi::connection::Clock> clock,
    std::string replay) :
    component(component), clock(std::move(clock)), activeReplay(std::move(replay))
{
}

armarx::viz::data::LayerUpdates
avi::connection::ReplayStorageAdapter::getLayerUpdates()
{
    try
    {
        auto time = std::chrono::duration_cast<std::chrono::microseconds>(
                        std::chrono::steady_clock::now() - last)
                        .count();

        {
            if(!buffering) {
                if(batchLoadQueue.size() < 4) {
                    clock->advance(time, false);
                    clock->setBuffering(false);
                } else {
                    clock->setBuffering(true);
                    buffering = true;
                }
            } else if(batchLoadQueue.size() < 2) {
                buffering = false;
            }
        }

        if(clock->hasRevisionUpdate()) {
            skipToRevision(clock->getTargetRevision());
        }

        int64_t index = -1;
        armarx::viz::data::LayerUpdates updates;
        int64_t wantedTimeStamp = clock->getTargetTime() + clock->getTimeFrame().begin;

        {   // get index of batch containing wantedTimeStamp
            if (wantedTimeStamp > clock->getTimeFrame().end ||
                wantedTimeStamp < clock->getTimeFrame().begin)
            {
                return updates;
            }

            for (auto& batch : batches)
            {
                if (batch.firstTimestamp > wantedTimeStamp)
                {
                    index = batch.batchIndex - 1;
                    break;
                }
            }
            if (index == -1)
            {
                index = static_cast<int64_t>(batches.size() - 1);
            }

            for(auto invalidBatch : invalidBatches) {
                if(invalidBatch == index) {
                    return updates;
                }
            }
        }

        updateBatchLoadQueue(index);
        updateBatchCache();
        asyncLoadBatch();

        if (batchCache.find(index) == batchCache.end())
        {
            last = std::chrono::steady_clock::now();
            return updates;
        }

        armarx::viz::data::RecordingBatch currentBatch = batchCache.at(index).data;
        if (isFirstBatch)
        {
            for (auto initialStateUpdate : currentBatch.initialState)
            {
                bool skipUpdate = false;

                for(auto update : currentBatch.updates) {
                    if (update.update.name == initialStateUpdate.update.name &&
                        update.update.component == initialStateUpdate.update.component)
                        skipUpdate = true;
                }

                if(!skipUpdate) updates.updates.push_back(initialStateUpdate.update);
            }
            isFirstBatch = false;
        }

        {   //push back revisions of batch
            batchCache.at(index).lastUsed = IceUtil::Time::now();
            size_t i =
                    binarySearch(currentBatch.updates, 0, currentBatch.updates.size(), wantedTimeStamp);
            clock->updateTargetRevision(currentBatch.updates[i].revision);
            for(;lastSentUpdateIndex < i; lastSentUpdateIndex++) {
                updates.updates.push_back(currentBatch.updates[lastSentUpdateIndex].update);
            }
        }

        last = std::chrono::steady_clock::now();
        return updates;
    }
    catch (IceUtil::Exception& e)
    {
        ARMARX_ERROR << "Error while pulling updates from storage: " << e.what();
        return {};
    }
}

size_t
avi::connection::ReplayStorageAdapter::binarySearch(
    const std::vector<armarx::viz::data::TimestampedLayerUpdate>& updates,
    size_t l,
    size_t r,
    int64_t timeStamp)
{
    if (r >= l)
    {
        size_t mid = l + (r - l) / 2;
        if ((updates[mid].timestampInMicroseconds <= timeStamp) &&
            (mid + 1 == updates.size() || updates[mid + 1].timestampInMicroseconds > timeStamp))
        {
            return mid;
        }
        else if (updates[mid].timestampInMicroseconds > timeStamp)
        {
            return binarySearch(updates, l, mid - 1, timeStamp);
        }
        else
        {
            return binarySearch(updates, mid + 1, r, timeStamp);
        }
    }
    else
    {
        return -1;
    }
}

std::vector<std::string>
avi::connection::ReplayStorageAdapter::getAllReplays()
{
    try
    {
        std::vector<armarx::viz::data::Recording> replays =
            component.getStorageInterface()->getAllRecordings().recordings;
        std::vector<std::string> replayNames;
        replayNames.reserve(replays.size());

        for (const auto& replay : replays)
        {
            replayNames.push_back(replay.id);
        }

        return replayNames;
    }
    catch (IceUtil::Exception& e)
    {
        ARMARX_ERROR << "Error while pulling replay names: " << e.what();
        return {};
    }
}

void
avi::connection::ReplayStorageAdapter::setup()
{
    try
    {
        auto recordings = component.getStorageInterface()->getAllRecordings().recordings;

        if (recordings.empty())
            return;

        size_t index = 0;
        isFirstBatch = true;
        while (recordings[index].id != activeReplay)
        {
            index++;
        }
        batchCount = recordings[index].batchHeaders.size();
        clock->setTimeFrame(TimeFrame(recordings[index].firstTimestampInMicroSeconds,
                                      recordings[index].lastTimestampInMicroSeconds));
        clock->setRevisionFrame(recordings[index].firstRevision, recordings[index].lastRevision);

        for (size_t i = 0; i < recordings[index].batchHeaders.size(); i++)
        {
            batches.push_back(batchInfo{recordings[index].batchHeaders[i].firstTimestampInMicroSeconds,
                                        recordings[index].batchHeaders[i].lastTimestampInMicroSeconds,
                                     recordings[index].batchHeaders[i].firstRevision,
                                        recordings[index].batchHeaders[i].lastRevision,
                                        static_cast<int64_t>(i)});
        }

        clock->setBuffering(true);
        TimestampedRecordingBatch timestampedRecordingBatch;
        timestampedRecordingBatch.data =
            component.getStorageInterface()->getRecordingBatch(activeReplay, 0);
        timestampedRecordingBatch.lastUsed = IceUtil::Time::now();

        batchCache.insert(std::make_pair<size_t, TimestampedRecordingBatch>(
            0, std::move(timestampedRecordingBatch)));
        clock->setAvailableTimeFrame(
            TimeFrame(recordings[index].batchHeaders[0].firstTimestampInMicroSeconds,
                      recordings[index].batchHeaders[0].lastTimestampInMicroSeconds));
        clock->setBuffering(false);

        callbackData = new ArVizWidgetBatchCallback();
        callbackData->this_ = this;
        callback = armarx::viz::newCallback_StorageInterface_getRecordingBatch(
            callbackData,
            &ArVizWidgetBatchCallback::onSuccess,
            &ArVizWidgetBatchCallback::onFailure);
    }

    catch (IceUtil::Exception& e)
    {
        ARMARX_ERROR << "Error while loading replay information: " << e.what();
    }
}

void
avi::connection::ReplayStorageAdapter::setInteraction(
    std::vector<avi::connection::Interaction> newInteractions)
{
}
void
avi::connection::ReplayStorageAdapter::asyncLoadBatch()
{
    if (batchLoadQueue.empty() || asyncBatchLoading) return;

    asyncBatchLoading = true;
    int64_t nextBatchIndex = batchLoadQueue.front();

    callbackResult = component.getStorageInterface()->begin_getRecordingBatch(
        activeReplay, nextBatchIndex, callback);
}
void
avi::connection::ReplayStorageAdapter::updateBatchLoadQueue(int index)
{
    if(lastRequestedBatchIndex == index) return;

    lastRequestedBatchIndex = index;
    lastSentUpdateIndex = 0;
    int64_t batchesSize = batches.size();
    int sign = 1;
    if (clock->getSpeed() < 0)
    {
        sign = -1;
    }

    for (int64_t i = -(sign * BATCHES_BACK); i != (sign * (BATCHES_FRONT + 1)); i = i + sign)
    {
        int64_t batchIndex = (batchesSize + ((index + i) % batchesSize)) % batchesSize;
        bool alreadyLoaded = false;
        for (int64_t entry : batchLoadQueue)
        {
            if (batchIndex == entry)
            {
                alreadyLoaded = true;
            }
        }

        for (auto& entry : batchCache)
        {
            if (batchIndex == entry.first)
            {
                alreadyLoaded = true;
            }
        }

        for(auto entry : invalidBatches) {
            if(batchIndex == entry) {
                alreadyLoaded = true;
            }
        }

        if (!alreadyLoaded)
        {
            batchLoadQueue.push_back(batchIndex);
        }
    }

    //remove outdated entries
    for (auto it = batchLoadQueue.begin(); it != batchLoadQueue.end(); it++)
    {
        if (std::abs(index - *it) > BATCH_CACHE_SIZE)
        {
            batchLoadQueue.erase(it--);
            continue;
        }

        if(0 < sign * index - sign * (*it)) {
            batchLoadQueue.erase(it--);
        }
    }
}


void
avi::connection::ReplayStorageAdapter::onGetBatchAsync(
    const armarx::viz::data::RecordingBatch& batch,
    long batchIndex)
{
    batchMutex.lock();

    asyncReceivedBatches.emplace_back(AsyncBatch{batch,batchIndex});
    asyncBatchLoading = false;


    batchMutex.unlock();
}

void avi::connection::ReplayStorageAdapter::updateBatchCache() {

    if(asyncReceivedBatches.empty()) return;


    std::vector<armarx::viz::data::RecordingBatch> receivedBatches;

    batchMutex.lock();
    auto iterator = asyncReceivedBatches.begin();
    while (iterator != asyncReceivedBatches.end())
    {
            int batchHeaderIndex = iterator.base()->batch.header.index;
            if(!(batchHeaderIndex < 0 || batchHeaderIndex >= batchCount)) {
                receivedBatches.emplace_back(iterator.base()->batch);
            } else {
                invalidBatches.emplace_back(iterator->batchIndex);

                godot::PoolStringArray poolStringArray;
                poolStringArray.push_back("Invalid batch found: " + godot::String(std::to_string(iterator->batchIndex).c_str()));
                poolStringArray.push_back("Check ArVizStorage warning for more information.");
                godot::OS::get_singleton()->execute("notify-send", poolStringArray, false);
            }
            iterator = asyncReceivedBatches.erase(iterator);
    }
    batchMutex.unlock();

    for(auto batch : receivedBatches) {
        TimestampedRecordingBatch batchEntry;

        batchEntry.data = batch;
        batchEntry.lastUsed = IceUtil::Time::now();

        batchCache.insert(std::make_pair<size_t, TimestampedRecordingBatch>(
                batchEntry.data.header.index, std::move(batchEntry)));
    }

    //remove old batches
    if (batchCache.size() <= BATCH_CACHE_SIZE)
    {
        return;
    }

    // Remove the entry with the oldest last used timestamp
    auto oldestIter = batchCache.begin();
    for (auto iter = batchCache.begin(); iter != batchCache.end(); ++iter)
    {
        TimestampedRecordingBatch& entry = iter->second;
        if (entry.lastUsed < oldestIter->second.lastUsed)
        {
            oldestIter = iter;
        }
    }

    batchCache.erase(oldestIter);
}

void avi::connection::ReplayStorageAdapter::skipToRevision(long revision) {
    long timestamp = -1;
    int batchIndex = -1;
    for(auto batch : batches) {
        if(batch.firstRevision <= revision && batch.lastRevision > revision) {
            batchIndex = batch.batchIndex;
            break;
        }
    }

    TimestampedRecordingBatch recordingBatch;
    bool found = false;
    for(auto batch : batchCache) {
        if(batch.second.data.header.index == batchIndex) {
            recordingBatch = batch.second;
            found = true;
            break;
        }
    }

    if(not found) {
        clock->setBuffering(true);
        recordingBatch.data = component.getStorageInterface()->getRecordingBatch(activeReplay, batchIndex);
        recordingBatch.lastUsed = IceUtil::Time::now();
        batchCache.insert(std::make_pair<size_t, TimestampedRecordingBatch>(
                batchIndex, std::move(recordingBatch)));
        clock->setBuffering(false);

        for(auto batch : batchCache) {
            if(batch.second.data.header.index == batchIndex) {
                recordingBatch = batch.second;
                break;
            }
        }
    }

    for(auto update : recordingBatch.data.updates) {
        if(update.revision == revision) {
            timestamp = update.timestampInMicroseconds;
        }
    }

    clock->advance(timestamp - clock->getTargetTime() - clock->getTimeFrame().begin, true);
}

bool avi::connection::ReplayStorageAdapter::finishedProcessing() {
    return !asyncBatchLoading;
}