#include "ModelLoader.h"

#include <math.h>

#include <SimoxUtility/algorithm/string/string_tools.h>
#include <VirtualRobot/ManipulationObject.h>
#include <VirtualRobot/Visualization/CoinVisualization/CoinVisualization.h>
#include <VirtualRobot/Visualization/TriMeshModel.h>
#include <VirtualRobot/Visualization/VisualizationNode.h>
#include <VirtualRobot/XML/ObjectIO.h>

#include <RobotAPI/libraries/ArmarXObjects/ObjectFinder.h>

#include <ArrayMesh.hpp>
#include <Godot.hpp>
#include <ImageTexture.hpp>
#include <ResourceLoader.hpp>
#include <arviz_godot/lib/avi/core/Constants.h>
#include <assimp/Importer.hpp>
#include <assimp/postprocess.h>
#include <assimp/scene.h>

avi::core::ModelLoader::ModelLoader() : modelCache()
{
    godot::ResourceLoader* loader = godot::ResourceLoader::get_singleton();
    godot::Ref<godot::Mesh> loadedErrorMesh = loader->load("res://models/error.obj");
    godot::SpatialMaterial* material = godot::SpatialMaterial::_new();
    material->set_flag(godot::SpatialMaterial::FLAG_UNSHADED, false);
    material->set_albedo(godot::Color(1, 0, 0, 1));

    {
        errorMesh = loadedErrorMesh.ptr();

        for (int64_t idx = 0; idx < errorMesh->get_surface_count(); idx++)
        {
            errorMesh->surface_set_material(idx, material);
        }
    }
}

godot::Ref<godot::Mesh>
avi::core::ModelLoader::loadModel(const std::string& name, std::string& fullName)
{
    std::string objectName = std::filesystem::path(name).filename().stem();
    armarx::ObjectFinder objectFinder;

    if (std::optional<armarx::ObjectInfo> info = objectFinder.findObject("", objectName))
    {
        fullName = info->simoxXML().absolutePath;
    }
    else
    {
        ARMARX_INFO << "Could not find object info for: '" << name << "'";
        fullName = name;
    }

    godot::Ref<godot::Mesh> returnObject = loadModelFromFile(fullName);

    return returnObject;
}

int64_t
convertMesh(const aiMesh* aiMesh, const std::string& name, godot::Ref<godot::ArrayMesh> arrayMesh)
{
    (void)name; // Can be used for debugging.

    bool hasNormals = aiMesh->HasNormals();
    bool hasTexture = aiMesh->HasTextureCoords(0);
    bool hasVertexColors = aiMesh->HasVertexColors(0);

    godot::PoolVector3Array coords;
    coords.resize(aiMesh->mNumFaces * 3);
    godot::Vector3* coordsPtr = coords.write().ptr();

    godot::PoolVector3Array normals;
    normals.resize(aiMesh->mNumFaces * 3);
    godot::Vector3* normalsPtr = normals.write().ptr();

    godot::PoolVector2Array textureCoords;
    if (hasTexture)
    {
        textureCoords.resize(aiMesh->mNumFaces * 3);
    }
    godot::Vector2* texturesPtr = textureCoords.write().ptr();

    godot::PoolColorArray colorArray;
    if (hasVertexColors)
    {
        colorArray.resize(aiMesh->mNumVertices);
    }
    godot::Color* colorsPtr = colorArray.write().ptr();

    for (size_t i = 0; i < aiMesh->mNumFaces; i++)
    {
        aiFace& face = aiMesh->mFaces[i];
        size_t v = i * 3;

        if (face.mNumIndices != 3)
        {
            continue;
        }

        aiVector3t<float>* vertex;
        vertex = &aiMesh->mVertices[face.mIndices[2]];
        coordsPtr[v + 0] = godot::Vector3(vertex->x, vertex->y, vertex->z);
        vertex = &aiMesh->mVertices[face.mIndices[1]];
        coordsPtr[v + 1] = godot::Vector3(vertex->x, vertex->y, vertex->z);
        vertex = &aiMesh->mVertices[face.mIndices[0]];
        coordsPtr[v + 2] = godot::Vector3(vertex->x, vertex->y, vertex->z);

        if (hasNormals)
        {
            aiVector3D* normal;
            normal = &aiMesh->mNormals[face.mIndices[2]];
            normalsPtr[v + 0] = godot::Vector3(normal->x, normal->y, normal->z);
            normal = &aiMesh->mNormals[face.mIndices[1]];
            normalsPtr[v + 1] = godot::Vector3(normal->x, normal->y, normal->z);
            normal = &aiMesh->mNormals[face.mIndices[0]];
            normalsPtr[v + 2] = godot::Vector3(normal->x, normal->y, normal->z);
        }
        else
        {
            vertex = &aiMesh->mVertices[face.mIndices[0]];
            godot::Vector3 vertA { vertex->x, vertex->y, vertex->z };
            vertex = &aiMesh->mVertices[face.mIndices[1]];
            godot::Vector3 vertB{vertex->x, vertex->y, vertex->z};
            vertex = &aiMesh->mVertices[face.mIndices[2]];
            godot::Vector3 vertC{vertex->x, vertex->y, vertex->z};

            godot::Vector3 vertAB{vertA.x - vertB.x, vertA.y - vertB.y, vertA.z - vertB.z};
            vertAB *= -1;
            godot::Vector3 vertAC{vertA.x - vertC.x, vertA.y - vertC.y, vertA.z - vertC.z};
            vertAC *= -1;
            godot::Vector3 orth = godot::vec3_cross(vertAB, vertAC);
            orth.normalize();

            normalsPtr[v + 0] = orth;
            normalsPtr[v + 1] = orth;
            normalsPtr[v + 2] = orth;
        }

        if (hasTexture)
        {
            aiVector3D* texture;
            texture = &aiMesh->mTextureCoords[0][face.mIndices[2]];
            texturesPtr[v + 0] = godot::Vector2(texture->x, texture->y);
            texture = &aiMesh->mTextureCoords[0][face.mIndices[1]];
            texturesPtr[v + 1] = godot::Vector2(texture->x, texture->y);
            texture = &aiMesh->mTextureCoords[0][face.mIndices[0]];
            texturesPtr[v + 2] = godot::Vector2(texture->x, texture->y);
        }
    }

    if (hasVertexColors)
    {
        aiColor4D* color = aiMesh->mColors[0];
        for (size_t i = 0; i < aiMesh->mNumVertices; i++)
        {
            aiColor4D current = color[i];
            colorsPtr[i] = godot::Color(current.r, current.g, current.b);
        }
    }

    godot::Array arr;
    arr.resize(godot::ArrayMesh::ARRAY_MAX);
    arr[godot::ArrayMesh::ARRAY_VERTEX] = coords;
    arr[godot::ArrayMesh::ARRAY_NORMAL] = normals;

    if (hasVertexColors)
    {
        arr[godot::ArrayMesh::ARRAY_COLOR] = colorArray;
    }

    if (hasTexture)
    {
        arr[godot::ArrayMesh::ARRAY_TEX_UV] = textureCoords;
    }

    int surface = arrayMesh->get_surface_count();

    arrayMesh->add_surface_from_arrays(godot::Mesh::PRIMITIVE_TRIANGLES, arr);

    return surface;
}

godot::Ref<godot::SpatialMaterial>
convertMaterial(const aiMesh* aiMesh, const aiScene* scene, const std::filesystem::path& path)
{
    godot::Ref<godot::SpatialMaterial> material = godot::SpatialMaterial::_new();
    godot::PoolByteArray bytes;

    material->set_flag(godot::SpatialMaterial::FLAG_UNSHADED, false);
    material->set_cull_mode(godot::SpatialMaterial::CULL_DISABLED);

    aiMaterial* mat = nullptr;
    if (aiMesh->mMaterialIndex < scene->mNumMaterials)
    {
        mat = scene->mMaterials[aiMesh->mMaterialIndex];
    }

    if (mat != nullptr)
    {
        aiColor3D diffuseColor(1.0, 1.0, 1.0);
        mat->Get(AI_MATKEY_COLOR_DIFFUSE, diffuseColor);
        material->set_albedo(godot::Color(diffuseColor.r, diffuseColor.g, diffuseColor.b));

        if (material->get_albedo().a < 1.0F)
        {
            material->set_feature(godot::SpatialMaterial::FEATURE_TRANSPARENT, true);
        }

        const float defaultSpecular = 0.5;
        aiColor3D specularColor(defaultSpecular, defaultSpecular, defaultSpecular);
        mat->Get(AI_MATKEY_COLOR_SPECULAR, specularColor);
        material->set_specular(specularColor.r);

        aiColor3D ambientColor(0.0, 0.0, 0.0);
        mat->Get(AI_MATKEY_COLOR_AMBIENT, ambientColor);
        material->set_metallic(ambientColor.r != 1.0 ? ambientColor.r : 0.0);

        float roughness = 0.5;
        if (mat->Get(AI_MATKEY_SHININESS, roughness) == AI_SUCCESS)
        {
            roughness /= 1000.0;
            roughness = std::sqrt(roughness);
            roughness = 1.0 - roughness;
        }

        material->set_roughness(roughness);
    }

    material->set_diffuse_mode(godot::SpatialMaterial::DIFFUSE_LAMBERT);

    aiString fileName;

    if (mat != nullptr)
    {
        aiGetMaterialTexture(mat, aiTextureType_DIFFUSE, 0, &fileName);
    }

    if (fileName != aiString())
    {
        std::string filePath = path.parent_path() / fileName.C_Str();

        FILE* file = fopen(filePath.c_str(), "rb");
        if (file == nullptr)
        {
            return nullptr;
        }

        // get length of file
        fseek(file, 0, SEEK_END);
        size_t length = ftell(file);
        rewind(file);
        bytes.resize(length);
        size_t read = fread(bytes.write().ptr(), 1, length, file);
        fclose(file);

        if (read != length)
        {
            ARMARX_WARNING << "Failed to load image '" << filePath << "'";
        }

        std::string imageExt = std::filesystem::path{filePath}.extension();

        godot::Ref<godot::Image> image = godot::Image::_new();

        if (imageExt == ".png")
        {
            image->load_png_from_buffer(bytes);
        }
        else if (imageExt == ".jpg")
        {
            image->load_jpg_from_buffer(bytes);
        }

        image->flip_y();

        godot::Ref<godot::ImageTexture> texture = godot::ImageTexture::_new();
        texture->create_from_image(image);

        material->set_texture(godot::SpatialMaterial::TEXTURE_ALBEDO, texture);
    }

    return material;
}


godot::Ref<godot::Mesh>
avi::core::ModelLoader::loadModelFromFile(const std::string& name)
{
    std::optional<godot::Ref<godot::Mesh>> out = modelCache.get(name);

    if (out)
    {
        return out.value();
    }

    std::string fullName = name;

    std::filesystem::path path{fullName};
    const std::string ext = simox::alg::to_lower(path.extension());

    if (ext != ".obj")
    {
        fullName = path.replace_extension(".obj");
    }

    godot::Ref<godot::Mesh> object;

    Assimp::Importer importer;
    const aiScene* scene =
        importer.ReadFile(fullName, aiProcess_RemoveRedundantMaterials | aiProcess_Triangulate);

    if (scene == nullptr)
    {
        object = loadModelFromNonObjFile(name);
    }
    else
    {
        if (scene->mNumMeshes == 0)
        {
            ARMARX_DEBUG << "Loading empty object '" << name << "'.";
        }

        godot::Ref<godot::ArrayMesh> mesh = godot::ArrayMesh::_new();

        for (size_t i = 0; i < scene->mNumMeshes; ++i)
        {
            aiMesh* aiMesh = scene->mMeshes[i];

            int64_t surface = convertMesh(aiMesh, name, mesh);

            godot::Ref<godot::SpatialMaterial> material = convertMaterial(aiMesh, scene, path);

            if (material.is_null())
            {
                ARMARX_ERROR << "Could not find object '" << name << "'.";
                return error();
            }

            material->set_flag(godot::SpatialMaterial::FLAG_ALBEDO_FROM_VERTEX_COLOR, true);

            mesh->surface_set_material(surface, material);
        }

        object = mesh;
    }

    modelCache.emplace(name, object);

    return object;
}

godot::Ref<godot::Mesh>
avi::core::ModelLoader::loadModelFromNonObjFile(const std::string& name)
{
    std::string fullName = name;
    std::filesystem::path path{fullName};
    const std::string ext = simox::alg::to_lower(path.extension());

    VirtualRobot::ManipulationObjectPtr object = getObject(name);
    VirtualRobot::TriMeshModelPtr model = object ? object->getVisualization()->getTriMeshModel() : nullptr;

    if (!model)
    {
        ARMARX_WARNING << "Error loading mesh: '" << name << "'";
        return error();
    }

    return createObjectFromTriMesh(model);
}

godot::Ref<godot::Mesh>
avi::core::ModelLoader::createObjectFromTriMesh(std::shared_ptr<VirtualRobot::TriMeshModel> model)
{
    const float scaleFactor = 1.0f / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR;
    godot::Ref<godot::ArrayMesh> mesh = godot::ArrayMesh::_new();
    godot::PoolVector3Array coords;
    godot::PoolVector3Array normals;
    godot::PoolColorArray colors;
    godot::PoolVector2Array textureCoords;

    Eigen::Matrix<float, 3, 1>* vertex;
    Eigen::Matrix<float, 3, 1>* normal;
    VirtualRobot::VisualizationFactory::Color* color;
    godot::Color colorGd;

    for (size_t i = 0; i < model->faces.size(); i++)
    {
        VirtualRobot::MathTools::TriangleFace& face = model->faces.at(i);
        vertex = &model->vertices.at(face.id1);
        coords.push_back(godot::Vector3(vertex->x(), vertex->y(), vertex->z()) * scaleFactor);
        vertex = &model->vertices.at(face.id2);
        coords.push_back(godot::Vector3(vertex->x(), vertex->y(), vertex->z()) * scaleFactor);
        vertex = &model->vertices.at(face.id3);
        coords.push_back(godot::Vector3(vertex->x(), vertex->y(), vertex->z()) * scaleFactor);

        normal = &face.normal;
        normals.push_back(godot::Vector3(normal->x(), normal->y(), normal->z()));
        normals.push_back(godot::Vector3(normal->x(), normal->y(), normal->z()));
        normals.push_back(godot::Vector3(normal->x(), normal->y(), normal->z()));

        if (!model->materials.empty())
        {
            color = &model->materials[face.idMaterial].diffuse;
            colorGd = godot::Color(color->r, color->g, color->b, color->transparency);
            colors.push_back(colorGd);
            colors.push_back(colorGd);
            colors.push_back(colorGd);
        }
    }

    godot::Array arr{};
    arr.resize(godot::ArrayMesh::ARRAY_MAX);
    arr[godot::ArrayMesh::ARRAY_VERTEX] = coords;
    if (normals.size() > 0)
    {
        arr[godot::ArrayMesh::ARRAY_NORMAL] = normals;
    }
    if (colors.size() > 0)
    {
        arr[godot::ArrayMesh::ARRAY_COLOR] = colors;
    }
    mesh->add_surface_from_arrays(godot::Mesh::PRIMITIVE_TRIANGLES, arr);

    godot::SpatialMaterial* material = godot::SpatialMaterial::_new();
    material->set_flag(godot::SpatialMaterial::FLAG_UNSHADED, false);
    material->set_cull_mode(godot::SpatialMaterial::CULL_DISABLED);

    mesh->surface_set_material(0, material);

    return mesh;
}

VirtualRobot::ManipulationObjectPtr avi::core::ModelLoader::getObject(const std::string& name)
{
    std::optional<VirtualRobot::ManipulationObjectPtr> object = objectCache.get(name);
    if (object.has_value())
        return object.value();

    std::filesystem::path path{name};
    const std::string ext = simox::alg::to_lower(path.extension());

    if (not std::filesystem::exists(path))
    {
        return nullptr; // Apparently the methods below do not return null when the file does not exist.
    }

    if (ext == ".xml" || ext == ".moxml")
    {
        VirtualRobot::ManipulationObjectPtr visPointer =
            VirtualRobot::ObjectIO::loadManipulationObject(name);

        objectCache.emplace(name, visPointer);

        return visPointer;
    }

    if (ext == ".wrl" || ext == ".iv")
    {
        VirtualRobot::VisualizationFactoryPtr factory =
            VirtualRobot::VisualizationFactory::fromName("inventor", nullptr);
        VirtualRobot::VisualizationNodePtr vis = factory->getVisualizationFromFile(name);
        VirtualRobot::ManipulationObjectPtr visPointer =
            std::make_shared<VirtualRobot::ManipulationObject>(name, vis);

        objectCache.emplace(name, visPointer);

        return visPointer;
    }

    return nullptr;
}

avi::core::Cache<godot::Mesh*, godot::Ref<godot::Shape>>&
avi::core::ModelLoader::getShapeCache()
{
    return shapeCache;
}

avi::core::Cache<VirtualRobot::ManipulationObjectPtr, godot::Ref<godot::Shape>>&
avi::core::ModelLoader::getCollisionShapeCache()
{
    return collisionShapeCache;
}

void
avi::core::ModelLoader::clearCaches()
{
    modelCache.clear();
    shapeCache.clear();
    collisionShapeCache.clear();
    objectCache.clear();
}

godot::Ref<godot::Mesh>
avi::core::ModelLoader::error()
{
    return errorMesh;
}
