#include "OptionFile.h"

#include <exception>
#include <fstream>
#include <utility>

#include <ArmarXCore/core/logging/Logging.h>

const std::string&
avi::core::OptionFile::getName()
{
    return name;
}

void
avi::core::OptionFile::setName(const std::string& name)
{
    this->name = name;
}

std::optional<avi::core::OptionFile>
avi::core::OptionFile::load(const std::filesystem::path& path)
{
    std::ifstream file(path);

    if (!file.is_open())
    {
        ARMARX_INFO << "Could not open settings file '" << path.string()
                    << "', using default values";
        return std::nullopt;
    }

    std::string name;
    std::map<std::string, std::string> options;

    bool first_line = true;

    for (std::string line; std::getline(file, line);)
    {
        if (line.empty())
        {
            continue;
        }

        if (first_line)
        {
            name = line;
            first_line = false;
            continue;
        }

        auto delimiter = line.find('=');

        if (delimiter == std::string::npos)
        {
            continue;
        }

        std::string key = line.substr(0, delimiter);
        std::string value = line.substr(delimiter + 1);

        options.emplace(key, value);
    }

    return OptionFile(name, options);
}

avi::core::OptionFile
avi::core::OptionFile::create(const std::string& name)
{
    return {name, std::map<std::string, std::string>()};
}

void
avi::core::OptionFile::save(const std::filesystem::path& path)
{
    std::ofstream file(path);

    if (!file.is_open())
    {
        ARMARX_ERROR << "Failed to save to file " << path;
        return;
    }

    file << name << std::endl;

    for (const auto& [key, value] : options)
    {
        file << key << "=" << value << std::endl;
    }
}

avi::core::OptionFile::OptionFile(std::string name, std::map<std::string, std::string> options) :
    name(std::move(name)), options(std::move(options))
{
}

#define SPECIALIZED_GET(TYPE)                                                                      \
    template <>                                                                                    \
    std::optional<TYPE> avi::core::OptionFile::getValue<TYPE>(const std::string& key)

#define SPECIALIZED_SET(TYPE)                                                                      \
    template <>                                                                                    \
    void avi::core::OptionFile::setValue<TYPE>(const std::string& key, TYPE value)

SPECIALIZED_GET(int64_t)
{
    auto value = getValue<std::string>(key);
    if (!value)
    {
        return std::nullopt;
    }

    try
    {
        return std::stoll(value.value());
    }
    catch (const std::exception& e)
    {
        try
        {
            return static_cast<int64_t>(std::stod(value.value()));
        }
        catch (const std::exception& e)
        {
            return std::nullopt;
        }
    }
}

SPECIALIZED_SET(int64_t)
{
    setValue<std::string>(key, std::to_string(value));
}

SPECIALIZED_GET(double)
{
    auto value = getValue<std::string>(key);
    if (!value)
    {
        return std::nullopt;
    }

    try
    {
        return std::stod(value.value());
    }
    catch (const std::exception& e)
    {
        return std::nullopt;
    }
}

SPECIALIZED_SET(double)
{
    setValue<std::string>(key, std::to_string(value));
}

SPECIALIZED_GET(std::string)
{
    auto it = options.find(key);

    if (it == options.end())
    {
        return std::nullopt;
    }

    return it->second;
}

SPECIALIZED_SET(std::string)
{
    options[key] = std::move(value);
}

SPECIALIZED_GET(bool)
{
    auto value = getValue<std::string>(key);
    if (!value)
    {
        return std::nullopt;
    }

    if (value.value() == "true")
    {
        return true;
    }

    if (value.value() == "false")
    {
        return false;
    }

    return std::nullopt;
}

SPECIALIZED_SET(bool)
{
    setValue<std::string>(key, value ? "true" : "false");
}

SPECIALIZED_GET(std::filesystem::path)
{
    auto value = getValue<std::string>(key);
    if (!value)
    {
        return std::nullopt;
    }

    return std::filesystem::path(value.value());
}

SPECIALIZED_SET(std::filesystem::path)
{
    setValue<std::string>(key, value.string());
}

SPECIALIZED_GET(godot::Color)
{
    auto value = getValue<std::string>(key);
    if (!value)
    {
        return std::nullopt;
    }

    godot::Color color = godot::Color::html(value.value().c_str());

    return color;
}

SPECIALIZED_SET(godot::Color)
{
    std::string html = value.to_html().ascii().get_data();
    setValue<std::string>(key, html);
}
