#include "UserInterface.h"

#include <Input.hpp>
#include <InputEventKey.hpp>
#include <MenuButton.hpp>
#include <PopupMenu.hpp>
#include <RichTextLabel.hpp>
#include <ShortCut.hpp>
#include <arviz_godot/lib/avi/gui/CameraModeMenu.h>
#include <arviz_godot/lib/avi/gui/SelectionMenu.h>
#include <arviz_godot/lib/avi/gui/StatusIcons.h>

void
avi::gui::UserInterface::_register_methods()
{
    register_method("_ready", &UserInterface::_ready);
    register_method("_scene_menu_select", &UserInterface::_scene_menu_select);
    register_method("_view_menu_select", &UserInterface::_view_menu_select);
}

void
avi::gui::UserInterface::_init()
{
}

void
avi::gui::UserInterface::_ready()
{
    layerSelection = get_node<avi::gui::LayerSelection>("Overlays/Layers");
    graphicsSettings = get_node<avi::gui::GraphicsSettings>("Overlays/Visuals");
    replayCockpit = get_node<avi::gui::ReplayCockpit>("Overlays/Replay");

    settingsWindow = get_node<avi::gui::SettingsWindow>("Settings");
    orientationWindow = get_node<avi::gui::OrientationWindow>("Orientation");

    get_node<godot::MenuButton>("Bar/Bar Background/Bar Container/Scene Menu")
        ->get_popup()
        ->connect("id_pressed", this, "_scene_menu_select");

    get_node<godot::MenuButton>("Bar/Bar Background/Bar Container/View Menu")
        ->get_popup()
        ->connect("id_pressed", this, "_view_menu_select");

    helpWindow = get_node<godot::Popup>("Help");
}

void
avi::gui::UserInterface::setupUI(avi::core::EventInvokingSystem& eventInvokingSystem,
                                 avi::core::EventSubscriptionSystem& eventSubscriptionSystem,
                                 avi::core::EventToKeyProvider& eventToKeyProvider,
                                 avi::core::Settings& settings)
{
    this->eventInvokingSystem = &eventInvokingSystem;
    this->eventSubscriptionSystem = &eventSubscriptionSystem;

    settingsWindow->setup(settings);

    setupConstantHelpBar(eventToKeyProvider);

    get_node<avi::gui::CameraModeMenu>("Bar/Bar Background/Bar Container/CameraMenu")
        ->setup(eventInvokingSystem, eventSubscriptionSystem, eventToKeyProvider);

    eventSubscriptionSystem.subscribe("HIDE_UI",
                                      [this]
                                      {
                                          if (hiddenToggle)
                                          {
                                              requestVisibleUI();
                                          }
                                          else
                                          {
                                              requestHiddenUI();
                                          }

                                          hiddenToggle = !hiddenToggle;
                                      });

    viewMenuOptions = {"VIEW_CAMERA_RESET",
                       "VIEW_CAMERA_FOCUS_SCENE",
                       "TAKE_SCREENSHOT",
                       "RECORD_VIDEO",
                       "HIDE_UI",
                       "VIEW_CAMERA_SWITCH_PROJECTION",
                       "PLACE_ON_X_NUMPAD",
                       "PLACE_ON_Y_NUMPAD",
                       "PLACE_ON_Z_NUMPAD"};

    auto* viewMenu = get_node<godot::MenuButton>("Bar/Bar Background/Bar Container/View Menu");
    viewMenu->set_disable_shortcuts(true);

    for (int index = 0; index < viewMenu->get_popup()->get_item_count(); index++)
    {
        std::string const& event = viewMenuOptions[index];
        std::optional<godot::InputEventKey*> key = eventToKeyProvider.getEventKey(event);

        if (key.has_value())
        {
            godot::Ref<godot::ShortCut> shortCut = godot::ShortCut::_new();
            shortCut->set_shortcut(key.value());

            viewMenu->get_popup()->set_item_shortcut(index, shortCut);
        }
    }
}

void
avi::gui::UserInterface::setupConstantHelpBar(avi::core::EventToKeyProvider& eventToKeyProvider)
{
    auto* label = get_node<godot::RichTextLabel>("Bar/Bar Background/Bar Container/Constant Hint");

    std::string text = "[center]";

    auto addHint = [&text, &eventToKeyProvider](const char* event, const std::string& description)
    {
        std::optional<godot::InputEventKey*> key = eventToKeyProvider.getEventKey(event);

        std::string hint;

        if (key.has_value())
        {
            hint += " [";
            hint += key.value()->as_text().ascii().get_data();
            hint += "] ";
        }

        hint += description;

        text += hint;
    };

    addHint("VIEW_CAMERA_FOCUS_POINT", "Focus");
    addHint("VIEW_CAMERA_FOCUS_SCENE", "View All");
    addHint("VIEW_CAMERA_RESET", "Reset View");
    addHint("HIDE_UI", "Toggle UI");
    addHint("TAKE_SCREENSHOT", "Capture Screenshot");
    addHint("RECORD_VIDEO", "Record Video");
    addHint("VIEW_CAMERA_LOCK", "Lock Mouse");

    text += "[/center]";

    label->set_use_bbcode(true);
    label->set_bbcode(text.c_str());
}

void
avi::gui::UserInterface::setupStatusIcons(avi::connection::Connectable& connectable,
                                          avi::connection::StorageModeSwitch& mode,
                                          avi::recording::VideoRecorder& recorder,
                                          avi::scene::GraphicsController& graphicsController,
                                          avi::core::EventToKeyProvider& eventToKeyProvider)
{
    this->connectable = &connectable;

    auto* statusIcons = godot::Object::cast_to<avi::gui::StatusIcons>(get_node("StatusBar"));
    statusIcons->setup(connectable, mode, recorder, graphicsController, eventToKeyProvider);

    advertiser = statusIcons;
}

void
avi::gui::UserInterface::setupScene(avi::scene::VisualizedScene& visualizedScene,
                                    avi::recording::VideoRecorder& videoRecorder,
                                    avi::recording::ImageRecorder& imageRecorder)
{
    this->videoRecorder = &videoRecorder;
    this->imageRecorder = &imageRecorder;
    layerSelection->setup(visualizedScene);
    orientationWindow->setup(visualizedScene);
}

void
avi::gui::UserInterface::setupGraphics(avi::scene::GraphicsController& graphicsController,
                                       avi::scene::ProfileStorage& profileStorage)
{
    graphicsSettings->setup(graphicsController, profileStorage);
}

void
avi::gui::UserInterface::setupInteraction(
    avi::interaction::InteractionDispatch& interactionDispatch)
{
    this->interactionDispatch = &interactionDispatch;
    this->selectionMenu = get_node<avi::gui::SelectionMenu>("InteractionPopupMenu");
}

void
avi::gui::UserInterface::setupReplay(avi::replay::ReplayController& replayController,
                                     avi::scene::VisualizedScene& visualizedScene)
{
    this->replayCockpit->setup(replayController, visualizedScene, this);
}

void
avi::gui::UserInterface::setupManipulation(avi::core::EventToKeyProvider& eventToKeyProvider)
{
    manipulationMenu =
        get_node<avi::gui::ManipulationMenu>("Bar/Bar Background/Bar Container/ManipulationMenu");
    manipulationMenu->setup(
        *interactionDispatch->getManipulator(), *eventSubscriptionSystem, eventToKeyProvider);
}

void
avi::gui::UserInterface::_scene_menu_select(int64_t id)
{
    switch(id) {
        case 0: // Open Screenshot Directory
        {
            godot::PoolStringArray poolStringArray;
            poolStringArray.push_back(imageRecorder->getLastScreenshotPath().c_str());

            godot::OS::get_singleton()->execute("nautilus", poolStringArray, false);
            break;
        }
        case 1: // Open Video Directory
        {
            godot::PoolStringArray poolStringArray;
            poolStringArray.push_back(videoRecorder->getLastRecordingPath().c_str());

            godot::OS::get_singleton()->execute("nautilus", poolStringArray, false);
            break;
        }
        case 2: // Orientation
            orientationWindow->open();
            break;
        case 3: // Settings
            settingsWindow->open();
            break;
        case 4: // Help
            helpWindow->popup_centered();
            break;
    }
}

void
avi::gui::UserInterface::_view_menu_select(int64_t id)
{
    if (id >= 0 && id < static_cast<int64_t>(viewMenuOptions.size()))
    {
        eventInvokingSystem->invoke(viewMenuOptions[id]);
    }
    else
    {
        godot::Godot::print_error(
            "Unsupported option in menu 'view'", __FUNCTION__, __FILE__, __LINE__);
    }
}

void
avi::gui::UserInterface::requestHiddenUI()
{
    hidingCounter++;

    if (hidingCounter > 0)
    {
        set_visible(false);
    }
}

void
avi::gui::UserInterface::requestVisibleUI()
{
    assert(hidingCounter > 0 && "More visible requests than hidden requests.");

    hidingCounter--;

    if (hidingCounter == 0)
    {
        set_visible(true);
    }
}

void
avi::gui::UserInterface::openSelectionMenu(godot::Vector2 mousePosition,
                                           std::vector<std::string> interactions,
                                           std::function<void(int64_t)> callback)
{
    this->selectionMenu->open(mousePosition, interactions, callback);
}

void
avi::gui::UserInterface::openTextInputPrompt()
{
}

void
avi::gui::UserInterface::clear()
{
    layerSelection->clear();
}

void
avi::gui::UserInterface::lockLayerSelection()
{
    layerSelection->lock();
}

void
avi::gui::UserInterface::unlockLayerSelection()
{
    layerSelection->unlock();
}

avi::core::ControlAdvertiser&
avi::gui::UserInterface::getAdvertiser()
{
    return *advertiser;
}
