#include "Manipulator.h"

#include <Quat.hpp>

#include <arviz_godot/lib/avi/interaction/InteractionDispatch.h>

avi::interaction::Manipulator::Manipulator(avi::interaction::InteractionDispatch& dispatch)
{
    this->dispatch = &dispatch;
    this->manipulationMode = ManipulationMode::NONE;
}

void
avi::interaction::Manipulator::setManipulationMode(ManipulationMode& mode)
{
    if (this->manipulationMode == mode)
        return;

    this->manipulationMode = mode;

    if (targetNode == nullptr)
        return;

    switch (manipulationMode)
    {
        case (ManipulationMode::TRANSLATE):
            deactivateAllGizmos();
            gizmoNodeParent->add_child(move);
            this->move->activate();
            this->move->targetElement(*targetNode);
            break;
        case (ManipulationMode::ROTATE):
            deactivateAllGizmos();
            gizmoNodeParent->add_child(rotate);
            this->rotate->activate();
            this->rotate->targetElement(*targetNode);
            break;
        case (ManipulationMode::SCALE):
            deactivateAllGizmos();
            gizmoNodeParent->add_child(scale);
            this->scale->activate();
            this->scale->targetElement(*targetNode);
            break;
        case (ManipulationMode::NONE):
            deactivateAllGizmos();
            gizmoNodeParent->add_child(none);
            this->none->activate();
            this->none->targetElement(*targetNode);
            break;
    }
}

ManipulationMode
avi::interaction::Manipulator::getManipulationMode()
{
    return this->manipulationMode;
}

void
avi::interaction::Manipulator::target(avi::scene::ElementNode& element)
{
    this->targetNode = &element;

    godot::Transform target = this->targetNode->getTransform();
    this->proxy->set_transform(target);

    switch (manipulationMode)
    {
        case (ManipulationMode::TRANSLATE):
            deactivateAllGizmos();
            gizmoNodeParent->add_child(move);
            this->move->targetElement(*targetNode);
            this->move->activate();
            break;
        case (ManipulationMode::ROTATE):
            deactivateAllGizmos();
            gizmoNodeParent->add_child(rotate);
            this->rotate->targetElement(*targetNode);
            this->rotate->activate();
            break;
        case (ManipulationMode::SCALE):
            deactivateAllGizmos();
            gizmoNodeParent->add_child(scale);
            this->scale->targetElement(*targetNode);
            this->scale->activate();
            break;
        case (ManipulationMode::NONE):
            deactivateAllGizmos();
            gizmoNodeParent->add_child(none);
            this->none->targetElement(*targetNode);
            this->none->activate();
            break;
    }

    this->didTransform = false;

    advertiser->activateAdvertisements(adID);
}

void
avi::interaction::Manipulator::untarget()
{
    if (this->targetNode)
    {
        this->targetNode->setMeshVisible(true);
        this->targetNode = nullptr;
    }

    deactivateAllGizmos();

    advertiser->deactivateAdvertisements(adID);
}

void
avi::interaction::Manipulator::setGizmos(avi::interaction::Move& move,
                                         avi::interaction::Rotate& rotate,
                                         avi::interaction::Scale& scale,
                                         avi::interaction::None& none)
{
    this->move = &move;
    this->rotate = &rotate;
    this->scale = &scale;
    this->none = &none;

    gizmoNodeParent = move.get_parent();
    proxy = gizmoNodeParent->get_node<godot::Spatial>("Proxy");

    move.setup(*this, proxy);
    rotate.setup(*this, proxy);
    scale.setup(*this, proxy);
    none.setup(*this, proxy);
    deactivateAllGizmos();
}

void
avi::interaction::Manipulator::deactivateAllGizmos()
{
    if (gizmoNodeParent->has_node("Move"))
    {
        gizmoNodeParent->remove_child(move);
    }

    if (gizmoNodeParent->has_node("Rotate"))
    {
        gizmoNodeParent->remove_child(rotate);
    }

    if (gizmoNodeParent->has_node("Scale"))
    {
        gizmoNodeParent->remove_child(scale);
    }

    if (gizmoNodeParent->has_node("None"))
    {
        gizmoNodeParent->remove_child(none);
    }

    this->move->deactivate();
    this->rotate->deactivate();
    this->scale->deactivate();
    this->none->deactivate();
}

void
avi::interaction::Manipulator::endManipulation()
{
    if (targetNode == nullptr)
        return;

    avi::connection::Interaction interaction =
        createInteraction(avi::connection::Interaction::TRANSFORMATION_END);

    this->didTransform = false;

    targetNode->createGhostElement(this->proxy->get_transform());
    this->dispatch->sendInteraction(interaction);
}

void
avi::interaction::Manipulator::sendManipulation()
{
    if (targetNode == nullptr)
        return;

    auto transformation = didTransform ? avi::connection::Interaction::TRANSFORMATION_DURING
                                       : avi::connection::Interaction::TRANSFORMATION_BEGIN;
    avi::connection::Interaction interaction = createInteraction(transformation);

    didTransform = true;

    targetNode->createGhostElement(this->proxy->get_transform());
    this->dispatch->sendManipulation(interaction);
}

void
avi::interaction::Manipulator::setGizmoTransform(godot::Spatial* gizmo,
                                                 godot::Transform global,
                                                 bool isLocal)
{
    godot::Transform proxyTransform = proxy->get_transform().orthonormalized();

    godot::Transform globalTransform;
    globalTransform.origin = global.orthonormalized().origin;

    godot::Transform gizmoTransform = isLocal ? proxyTransform : globalTransform;

    gizmoTransform.origin = proxyTransform.origin;
    gizmo->set_transform(gizmoTransform);

    scaleGizmo(gizmo);
}

void
avi::interaction::Manipulator::scaleGizmo(godot::Spatial* gizmo)
{
    auto* scaler = godot::Object::cast_to<godot::Spatial>(gizmo->get_child(0));

    godot::AABB aabb = targetNode->getVolume();
    godot::Vector3 boxSize = aabb.size;

    godot::Vector3 targetScale = targetNode->getTransform().basis.get_scale();
    godot::Vector3 ghostScale = proxy->get_transform().basis.get_scale();
    godot::Vector3 factor = ghostScale / targetScale;

    godot::Vector3 size = boxSize * factor * 2.0f;
    float max = size[size.max_axis()];

    godot::Vector3 gizmoScale = godot::Vector3::ONE;
    if (max > 0.0001f)
        gizmoScale *= max;

    scaler->set_scale(gizmoScale);
}

avi::connection::Interaction
avi::interaction::Manipulator::createInteraction(
    avi::connection::Interaction::Transformation transformation)
{
    godot::Vector3 manipulatedScale = this->proxy->get_transform().basis.get_scale();

    godot::Transform unscaledTransform = this->proxy->get_transform();
    unscaledTransform.orthonormalize();

    godot::Transform change =
        unscaledTransform * this->targetNode->getTransform().orthonormalized().inverse();

    auto quaternion = godot::Quat(change.get_basis());
    godot::Vector3 position = change.origin;

    avi::connection::Interaction interaction = avi::connection::Interaction::createManipulation(
        targetNode, position, quaternion, manipulatedScale, transformation);

    return interaction;
}

void
avi::interaction::Manipulator::preDelete()
{
    move->queue_free();
    rotate->queue_free();
    scale->queue_free();
    none->queue_free();
}

void
avi::interaction::Manipulator::setupControlAdvertisement(avi::core::ControlAdvertiser& advertiser)
{
    this->advertiser = &advertiser;

    using Ad = avi::core::ControlAdvertiser::Advertisement;

    std::vector<avi::core::ControlAdvertiser::Advertisement> ads;
    ads.emplace_back(Ad::fromString("Manipulation -"));
    ads.emplace_back(Ad::fromEvent("None", "MANIPULATOR_NONE"));
    ads.emplace_back(Ad::fromEvent("Move", "MANIPULATOR_MOVE"));
    ads.emplace_back(Ad::fromEvent("Rotate", "MANIPULATOR_ROTATE"));
    ads.emplace_back(Ad::fromEvent("Scale", "MANIPULATOR_SCALE"));
    ads.emplace_back(Ad::fromEvent("Precise Snap", "MANIPULATE_PRECISE"));
    ads.emplace_back(Ad::fromEvent("Very Precise Snap", "MANIPULATE_VERY_PRECISE"));

    this->adID = advertiser.registerAdvertisement(std::move(ads));
}
