#include "Converter.h"

#include <ArrayMesh.hpp>
#include <TriangleMesh.hpp>
#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/scene/ArrowCircleElementNode.h>
#include <arviz_godot/lib/avi/scene/ArrowElementNode.h>
#include <arviz_godot/lib/avi/scene/CubeElementNode.h>
#include <arviz_godot/lib/avi/scene/CustomMeshElementNode.h>
#include <arviz_godot/lib/avi/scene/CylinderElementNode.h>
#include <arviz_godot/lib/avi/scene/CylindroidElementNode.h>
#include <arviz_godot/lib/avi/scene/EllipsoidElementNode.h>
#include <arviz_godot/lib/avi/scene/LineElementNode.h>
#include <arviz_godot/lib/avi/scene/MeshElementNode.h>
#include <arviz_godot/lib/avi/scene/ObjectElementNode.h>
#include <arviz_godot/lib/avi/scene/PathElementNode.h>
#include <arviz_godot/lib/avi/scene/PointCloudElementNode.h>
#include <arviz_godot/lib/avi/scene/PolygonElementNode.h>
#include <arviz_godot/lib/avi/scene/PoseElementNode.h>
#include <arviz_godot/lib/avi/scene/RobotElementNode.h>
#include <arviz_godot/lib/avi/scene/SphereElementNode.h>
#include <arviz_godot/lib/avi/scene/TextElementNode.h>

avi::scene::Converter::Converter(godot::Spatial* sceneRoot,
                                 avi::connection::RevisionBuffer& buffer,
                                 avi::scene::GraphicsController* graphicsController) :
    sceneRoot(sceneRoot), buffer(buffer), graphicsController(graphicsController)
{
}

std::map<avi::scene::LayerName, avi::scene::LayerNode*>
avi::scene::Converter::apply()
{
    for (const auto& layer : buffer.getChangedLayers())
    {
        std::optional<armarx::viz::data::LayerUpdate> optLayerUpdate = buffer.getUpdate(layer);
        if (optLayerUpdate.has_value())
        {
            armarx::viz::data::LayerUpdate layerUpdate = optLayerUpdate.value();
            LayerName name = {layerUpdate.component, layerUpdate.name};

            switch (layerUpdate.action)
            {
                case armarx::viz::data::Layer_CREATE_OR_UPDATE:
                {
                    std::optional<avi::scene::LayerNode*> layerNode;

                    if (layers.find(name) != layers.end())
                    {
                        layerNode = layers[name];
                    }

                    updateLayer(layerNode, layerUpdate);
                    break;
                }
                case armarx::viz::data::Layer_DELETE:
                    if (layers.find(name) != layers.end())
                    {
                        layers.at(name)->saveRemove();
                        layers.erase(name);
                    }
                    break;
            }
        }
    }

    return layers;
}

void
avi::scene::Converter::updateLayer(std::optional<avi::scene::LayerNode*> optLayer,
                                   armarx::viz::data::LayerUpdate update)
{
    std::set<avi::scene::ElementNode*> marked;
    avi::scene::LayerNode* layer;

    if (!optLayer.has_value())
    {
        layer = avi::scene::LayerNode::_new();
        sceneRoot->add_child(layer);

        layer->setName(update.component, update.name);

        layers.insert(
            std::pair<LayerName, avi::scene::LayerNode*>({update.component, update.name}, layer));
    }
    else
    {
        layer = optLayer.value();
    }

    for (auto& element : update.elements)
    {
        avi::scene::ElementNode* node = layer->getElement(element->id);

        if (node != nullptr)
        {
            if (!node->is_queued_for_deletion())
            {
                node = updateElement(layer, node, element);
            }
        }
        else
        {
            node = createElement(layer, element);
        }

        if (node != nullptr)
        {
            marked.insert(node);
        }
    }

    std::vector<avi::scene::ElementNode*> markedDelete;
    for (int i = 0; i < layer->get_child_count(); i++)
    {
        auto* child = godot::Object::cast_to<avi::scene::ElementNode>(layer->get_child(i));
        if (marked.find(child) == marked.end())
        {
            if (!child->is_queued_for_deletion())
            {
                markedDelete.push_back(child);
            }
        }
    }

    for (auto& orphan : markedDelete)
    {
        layer->removeElement(orphan->getName());
        orphan->queue_free();
    }
}

avi::scene::ElementNode*
avi::scene::Converter::updateElement(avi::scene::LayerNode* layer,
                                     avi::scene::ElementNode* element,
                                     IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    avi::connection::SentElementType type = getType(elementInfo);

    if (element->getType() != type)
    {
        layer->removeElement(element->getName());
        return createElement(layer, elementInfo);
    }

    element->update(elementInfo);
    return element;
}

avi::scene::ElementNode*
avi::scene::Converter::createElement(avi::scene::LayerNode* layer,
                                     IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    avi::scene::ElementNode* created = nullptr;
    avi::connection::SentElementType type = getType(elementInfo);

    switch (type)
    {
        case connection::SentElementType::BOX:
        {
            avi::scene::CubeElementNode* boxNode =
                avi::scene::CubeElementNode::newCubeElementNode(elementInfo->id, layer);

            created = boxNode;

            break;
        }

        case connection::SentElementType::SPHERE:
        {
            avi::scene::MeshElementNode* sphereNode =
                avi::scene::SphereElementNode::newSphereElementNode(elementInfo->id, layer);

            created = sphereNode;

            break;
        }

        case connection::SentElementType::CYLINDER:
        {
            avi::scene::MeshElementNode* cylinderNode =
                avi::scene::CylinderElementNode::newCylinderElementNode(elementInfo->id, layer);

            created = cylinderNode;

            break;
        }

        case connection::SentElementType::ELLIPSOID:
        {
            avi::scene::MeshElementNode* ellipsoidNode =
                avi::scene::EllipsoidElementNode::newEllipsoidElementNode(elementInfo->id, layer);

            created = ellipsoidNode;

            break;
        }

        case connection::SentElementType::CYLINDROID:
        {
            auto cylindroid =
                IceInternal::Handle<armarx::viz::data::ElementCylindroid>::dynamicCast(elementInfo);
            avi::scene::CylindroidElementNode* cylindroidNode =
                avi::scene::CylindroidElementNode::newCylindroidElementNode(cylindroid->id, layer);

            created = cylindroidNode;

            break;
        }

        case connection::SentElementType::LINE:
        {
            avi::scene::LineElementNode* lineNode =
                avi::scene::LineElementNode::newLineElementNode(elementInfo->id, layer);

            created = lineNode;
            break;
        }

        case connection::SentElementType::PATH:
        {
            avi::scene::PathElementNode* pathNode =
                avi::scene::PathElementNode::newPathElementNode(elementInfo->id, layer);

            created = pathNode;
            break;
        }

        case connection::SentElementType::POLYGON:
        {
            godot::ArrayMesh* mesh1 = getNonEmptyArrayMesh();
            godot::ArrayMesh* mesh2 = getNonEmptyArrayMesh();
            avi::scene::PolygonElementNode* polygonNode =
                avi::scene::PolygonElementNode::newPolygonElementNode(
                    {mesh1, mesh2},
                    avi::connection::SentElementType::POLYGON,
                    elementInfo->id,
                    layer);

            created = polygonNode;
            break;
        }

        case connection::SentElementType::MESH:
        {
            godot::ArrayMesh* mesh = getNonEmptyArrayMesh();
            avi::scene::CustomMeshElementNode* meshNode =
                avi::scene::CustomMeshElementNode::newCustomMeshElementNode(
                    mesh, avi::connection::SentElementType::MESH, elementInfo->id, layer);

            created = meshNode;

            break;
        }

        case avi::connection::SentElementType::TEXT:
        {
            avi::scene::TextElementNode* textNode =
                avi::scene::TextElementNode::newTextElementNode(sceneRoot);
            textNode->setName(elementInfo->id);
            textNode->setStartAttributes(
                avi::connection::SentElementType::TEXT, elementInfo->id, layer);
            created = textNode;
            break;
        }

        case avi::connection::SentElementType::ARROW:
        {
            auto arrow =
                IceInternal::Handle<armarx::viz::data::ElementArrow>::dynamicCast(elementInfo);
            avi::scene::ArrowElementNode* arrowNode =
                avi::scene::ArrowElementNode::newArrowElementNode();
            arrowNode->setAllAttributes(avi::connection::SentElementType::ARROW, arrow->id, layer);
            created = arrowNode;
            break;
        }

        case avi::connection::SentElementType::OBJECT:
        {
            auto object =
                IceInternal::Handle<armarx::viz::data::ElementObject>::dynamicCast(elementInfo);
            avi::scene::ObjectElementNode* objectNode =
                avi::scene::ObjectElementNode::newObjectElementNode(
                    &modelLoader, avi::connection::SentElementType::OBJECT, object->id, layer);

            created = objectNode;
            break;
        }

        case avi::connection::SentElementType::ARROWCIRCLE:
        {
            avi::scene::ArrowCircleElementNode* arrowCircleNode =
                avi::scene::ArrowCircleElementNode::newArrowCircleElementNode(
                    avi::connection::SentElementType::ARROWCIRCLE, elementInfo->id, layer);
            created = arrowCircleNode;
            break;
        }

        case avi::connection::SentElementType::POSE:
        {
            avi::scene::PoseElementNode* poseNode = avi::scene::PoseElementNode::newPoseElementNode(
                avi::connection::SentElementType::POSE, elementInfo->id, layer);
            created = poseNode;
            break;
        }

        case avi::connection::SentElementType::ROBOT:
        {
            auto robot =
                IceInternal::Handle<armarx::viz::data::ElementRobot>::dynamicCast(elementInfo);
            avi::scene::RobotElementNode* robotNode =
                avi::scene::RobotElementNode::newRobotElementNode();

            robotNode->setStartAttributes(graphicsController,
                                          &modelLoader,
                                          avi::connection::SentElementType::ROBOT,
                                          robot->id,
                                          layer,
                                          robot->interaction.enableFlags,
                                          robot->interaction.contextMenuOptions,
                                          getPosition(robot->pose),
                                          getScale(robot->scale),
                                          getRotation(robot->pose),
                                          getColor(robot->color),
                                          robot->project,
                                          robot->filename,
                                          ModelDrawStyle(robot->drawStyle),
                                          robot->jointValues);
            created = robotNode;
            break;
        }

        case avi::connection::SentElementType::POINT_CLOUD:
        {
            avi::scene::PointCloudElementNode* pointCloudNode =
                avi::scene::PointCloudElementNode::newPointCloudElementNode(
                    avi::connection::SentElementType::POINT_CLOUD, elementInfo->id, layer);

            created = pointCloudNode;
            break;
        }

        case connection::SentElementType::ERROR:
            assert(false && "unreachable");
            break;
    }

    layer->addElement(created);
    created->update(elementInfo);

    return created;
}

godot::Vector3
avi::scene::Converter::getPosition(armarx::viz::data::GlobalPose pose)
{
    return {adjustSize(pose.x), adjustSize(pose.y), adjustSize(pose.z)};
}

godot::Vector3
avi::scene::Converter::getScale(armarx::Vector3f vec)
{
    armarx::Vector3f adjVec = adjustSizeVector(vec);
    return {adjVec.e0, adjVec.e1, adjVec.e2};
}

godot::Quat
avi::scene::Converter::getRotation(armarx::viz::data::GlobalPose pose)
{
    return {pose.qx, pose.qy, pose.qz, pose.qw};
}

godot::Color
avi::scene::Converter::getColor(armarx::viz::data::Color color)
{
    return {color.r / 255.0f, color.g / 255.0f, color.b / 255.0f, color.a / 255.0f};
}

avi::connection::SentElementType
avi::scene::Converter::getType(IceInternal::Handle<armarx::viz::data::Element> element)
{
    if (IceInternal::Handle<armarx::viz::data::ElementArrow>::dynamicCast(element))
    {
        return avi::connection::SentElementType::ARROW;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementArrowCircle>::dynamicCast(element))
    {
        return avi::connection::SentElementType::ARROWCIRCLE;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementBox>::dynamicCast(element))
    {
        return avi::connection::SentElementType::BOX;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementCylinder>::dynamicCast(element))
    {
        return avi::connection::SentElementType::CYLINDER;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementCylindroid>::dynamicCast(element))
    {
        return avi::connection::SentElementType::CYLINDROID;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementEllipsoid>::dynamicCast(element))
    {
        return avi::connection::SentElementType::ELLIPSOID;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementLine>::dynamicCast(element))
    {
        return avi::connection::SentElementType::LINE;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementMesh>::dynamicCast(element))
    {
        return avi::connection::SentElementType::MESH;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementObject>::dynamicCast(element))
    {
        return avi::connection::SentElementType::OBJECT;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementPath>::dynamicCast(element))
    {
        return avi::connection::SentElementType::PATH;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementPointCloud>::dynamicCast(element))
    {
        return avi::connection::SentElementType::POINT_CLOUD;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementPolygon>::dynamicCast(element))
    {
        return avi::connection::SentElementType::POLYGON;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementPose>::dynamicCast(element))
    {
        return avi::connection::SentElementType::POSE;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementRobot>::dynamicCast(element))
    {
        return avi::connection::SentElementType::ROBOT;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementSphere>::dynamicCast(element))
    {
        return avi::connection::SentElementType::SPHERE;
    }

    if (IceInternal::Handle<armarx::viz::data::ElementText>::dynamicCast(element))
    {
        return avi::connection::SentElementType::TEXT;
    }

    assert(false && "Not Implemented.");
    return {};
}

float
avi::scene::Converter::adjustSize(float size)
{
    return size / avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR;
}

armarx::Vector3f
avi::scene::Converter::adjustSizeVector(armarx::Vector3f sizevec)
{
    armarx::Vector3f adjusted{};
    adjusted.e0 = adjustSize(sizevec.e0);
    adjusted.e1 = adjustSize(sizevec.e1);
    adjusted.e2 = adjustSize(sizevec.e2);
    return adjusted;
}

godot::Vector3
avi::scene::Converter::getNormal(armarx::Vector3f A, armarx::Vector3f B, armarx::Vector3f C)
{
    godot::PoolVector3Array coords;

    godot::Vector3 vertAB = godot::Vector3(B.e0 - A.e0, B.e1 - A.e1, B.e2 - A.e2);
    godot::Vector3 vertAC = godot::Vector3(C.e0 - A.e0, C.e1 - A.e1, C.e2 - A.e2);
    godot::Vector3 orth = godot::vec3_cross(vertAB, vertAC);
    orth.normalize();
    return orth;
}

void
avi::scene::Converter::clear()
{
    for (auto& [name, l] : layers)
    {
        l->saveRemove();
    }
    layers.clear();

    buffer.clearBuffer();
}

void
avi::scene::Converter::clearCaches()
{
    modelLoader.clearCaches();
}

godot::ArrayMesh*
avi::scene::Converter::getNonEmptyArrayMesh()
{
    godot::ArrayMesh* mesh = godot::ArrayMesh::_new();
    godot::PoolVector3Array coords;
    coords.push_back(godot::Vector3(0, 0, 0));
    godot::Array arr{};
    arr.resize(godot::ArrayMesh::ARRAY_MAX);
    arr[godot::ArrayMesh::ARRAY_VERTEX] = coords;
    mesh->add_surface_from_arrays(godot::Mesh::PRIMITIVE_POINTS, arr);
    return mesh;
}