#include "PathElementNode.h"

#include <CollisionShape.hpp>
#include <CylinderMesh.hpp>
#include <CylinderShape.hpp>
#include <MultiMesh.hpp>
#include <MultiMeshInstance.hpp>
#include <SphereMesh.hpp>
#include <StaticBody.hpp>
#include <arviz_godot/lib/avi/core/Constants.h>
#include <arviz_godot/lib/avi/core/Utilities.h>
#include <arviz_godot/lib/avi/scene/Converter.h>

void
avi::scene::PathElementNode::_register_methods()
{
    register_method("_ready", &PathElementNode::_ready);
}

void
avi::scene::PathElementNode::_ready()
{
}

void
avi::scene::PathElementNode::_init()
{
}

void
avi::scene::PathElementNode::update(IceInternal::Handle<armarx::viz::data::Element> elementInfo)
{
    auto path = IceInternal::Handle<armarx::viz::data::ElementPath>::dynamicCast(elementInfo);

    bool allPointsApproxEqual = points.size() == path->points.size();

    std::vector<godot::Vector3> newPoints;
    newPoints.reserve(path->points.size());
    for (size_t i = 0; i < path->points.size(); ++i)
    {
        godot::Vector3 point = {path->points[i].e0, path->points[i].e1, path->points[i].e2};
        newPoints.push_back(point);

        if (allPointsApproxEqual && i < points.size())
        {
            allPointsApproxEqual &= avi::core::utilities::is_equal_approx(points[i], point);
        }
    }

    if (!godot::Math::is_equal_approx(radius, path->lineWidth) || !allPointsApproxEqual)
    {
        radius = path->lineWidth;
        points = std::move(newPoints);

        buildPath();
    }

    updateAttributes(path->interaction.enableFlags,
                     path->interaction.contextMenuOptions,
                     avi::scene::Converter::getPosition(path->pose),
                     avi::scene::Converter::getScale(path->scale),
                     avi::scene::Converter::getRotation(path->pose),
                     avi::scene::Converter::getColor(path->color));
}

avi::scene::PathElementNode*
avi::scene::PathElementNode::newPathElementNode(std::string name, avi::core::Layer* layer)
{
    PathElementNode* node = PathElementNode::_new();
    node->setStartAttributes(name, layer);
    return node;
}

void
avi::scene::PathElementNode::setStartAttributes(std::string name, avi::core::Layer* layer)
{
    setType(avi::connection::SentElementType::PATH);
    setName(name);
    setLayer(layer);

    staticBody = godot::StaticBody::_new();
    this->add_child(staticBody);

    godot::MultiMeshInstance* lines = godot::MultiMeshInstance::_new();
    staticBody->add_child(lines);
    spatials.push_back(lines);

    meshLine = godot::MultiMesh::_new();
    meshLine->set_mesh(godot::CylinderMesh::_new());
    lines->set_multimesh(meshLine);

    meshLine->set_transform_format(godot::MultiMesh::TRANSFORM_3D);
    meshLine->set_color_format(godot::MultiMesh::COLOR_NONE);
    meshLine->set_custom_data_format(godot::MultiMesh::CUSTOM_DATA_NONE);

    godot::MultiMeshInstance* spheres = godot::MultiMeshInstance::_new();
    staticBody->add_child(spheres);
    spatials.push_back(spheres);

    meshSphere = godot::MultiMesh::_new();
    meshSphere->set_mesh(godot::SphereMesh::_new());
    spheres->set_multimesh(meshSphere);

    meshSphere->set_transform_format(godot::MultiMesh::TRANSFORM_3D);
    meshSphere->set_color_format(godot::MultiMesh::COLOR_NONE);
    meshSphere->set_custom_data_format(godot::MultiMesh::CUSTOM_DATA_NONE);

    material = godot::SpatialMaterial::_new();
    lines->set_material_override(material);
    spheres->set_material_override(material);
}

void
avi::scene::PathElementNode::updateAttributes(int flag,
                                              std::vector<std::string> interactions,
                                              godot::Vector3 position,
                                              godot::Vector3 scale,
                                              godot::Quat rotation,
                                              godot::Color color)
{
    setInteractionFlag(flag);
    setInteractions(std::move(interactions));

    setPosition(position);
    setRotation(rotation);
    setScale(scale);
    setColor(color);
}

godot::AABB
avi::scene::PathElementNode::getVolume()
{
    bool noBaseAABB = true;
    godot::AABB AABB;

    godot::Vector3 pathPosition = staticBody->get_global_transform().origin;
    godot::Vector3 pathScale = staticBody->get_scale();
    godot::Basis changeOfBasisMatrix = this->get_global_transform().basis;

    for (auto point : points)
    {
        if (noBaseAABB)
        {
            AABB.position = changeOfBasisMatrix.xform(point) * pathScale + pathPosition;
            noBaseAABB = false;
        }
        else
        {
            godot::AABB tempAABB;
            tempAABB.position = changeOfBasisMatrix.xform(point) * pathScale + pathPosition;
            AABB.merge_with(tempAABB);
        }
    }

    return AABB;
}

godot::Transform
avi::scene::PathElementNode::getTransform()
{
    godot::Transform transform = staticBody->get_transform();
    transform.basis.scale(godot::Vector3::ONE * avi::core::GODOT_TO_ARMARX_CONVERSION_FACTOR);

    return transform;
}

std::vector<godot::Spatial*>
avi::scene::PathElementNode::getSpatial()
{
    return spatials;
}

void
avi::scene::PathElementNode::buildPath()
{
    buildLines();
    buildSpheres();
}

void
avi::scene::PathElementNode::buildLines()
{
    const size_t perInstanceData = 12;
    size_t lineCount = points.size() - 1;
    lineCount = !points.empty() ? lineCount : 0;

    int requiredSize = static_cast<int>(lineCount * perInstanceData);
    if (lineData.size() != requiredSize)
    {
        lineData.resize(requiredSize);
    }

    godot::Vector3 p0 = !points.empty() ? points.front() : godot::Vector3::ZERO;
    minPoint = p0;
    maxPoint = p0;

    real_t* data = lineData.write().ptr();

    std::vector<godot::CollisionShape*> newShapes;
    newShapes.reserve(lineCount);
    godot::Shape* shape = godot::CylinderShape::_new();

    for (ulong lineIndex = 0; lineIndex < lineCount; lineIndex++)
    {
        godot::Vector3& a = points[lineIndex];
        godot::Vector3& b = points[lineIndex + 1];
        real_t* currentData = data + (lineIndex * perInstanceData);

        size_t i = 0;

        godot::Vector3 direction = b - a;
        godot::Vector3 mid = a + direction / 2;

        godot::Transform rotate;
        rotate.basis *= godot::Quat(godot::Vector3::UP, direction.normalized());

        godot::Transform scale;
        scale.scale({radius, direction.length() / 2, radius});

        godot::Basis basis = (rotate * scale).basis;

        currentData[i++] = basis.elements[0][0];
        currentData[i++] = basis.elements[0][1];
        currentData[i++] = basis.elements[0][2];
        currentData[i++] = mid.x;

        currentData[i++] = basis.elements[1][0];
        currentData[i++] = basis.elements[1][1];
        currentData[i++] = basis.elements[1][2];
        currentData[i++] = mid.y;

        currentData[i++] = basis.elements[2][0];
        currentData[i++] = basis.elements[2][1];
        currentData[i++] = basis.elements[2][2];
        currentData[i++] = mid.z;

        auto considerPointBounds = [&](const godot::Vector3& point)
        {
            minPoint = {std::min(minPoint.x, point.x),
                        std::min(minPoint.y, point.y),
                        std::min(minPoint.z, point.z)};

            maxPoint = {std::max(maxPoint.x, point.x),
                        std::max(maxPoint.y, point.y),
                        std::max(maxPoint.z, point.z)};
        };

        considerPointBounds(a);
        considerPointBounds(b);

        godot::CollisionShape* current;

        if (lineIndex < shapes.size())
        {
            current = shapes[lineIndex];
        }
        else
        {
            current = godot::CollisionShape::_new();
            staticBody->add_child(current);
            current->set_shape(shape);
        }

        current->set_transform(rotate * scale);
        current->set_translation(mid);

        newShapes.push_back(current);
    }

    meshLine->set_instance_count(static_cast<int64_t>(lineCount));
    meshLine->set_as_bulk_array(lineData);

    for (size_t index = newShapes.size(); index < shapes.size(); index++)
    {
        shapes[index]->queue_free();
    }

    shapes = std::move(newShapes);
}

void
avi::scene::PathElementNode::buildSpheres()
{
    const size_t perInstanceData = 12;
    size_t sphereCount = points.size();

    int requiredSize = static_cast<int>(sphereCount * perInstanceData);
    if (sphereData.size() != requiredSize)
    {
        sphereData.resize(requiredSize);
    }

    real_t* data = sphereData.write().ptr();

    for (ulong sphereIndex = 0; sphereIndex < sphereCount; sphereIndex++)
    {
        godot::Vector3& point = points[sphereIndex];
        real_t* currentData = data + (sphereIndex * perInstanceData);

        size_t i = 0;

        godot::Basis basis;
        basis.scale(godot::Vector3(radius, radius, radius) * 1.5f);

        currentData[i++] = basis.elements[0][0];
        currentData[i++] = basis.elements[0][1];
        currentData[i++] = basis.elements[0][2];
        currentData[i++] = point.x;

        currentData[i++] = basis.elements[1][0];
        currentData[i++] = basis.elements[1][1];
        currentData[i++] = basis.elements[1][2];
        currentData[i++] = point.y;

        currentData[i++] = basis.elements[2][0];
        currentData[i++] = basis.elements[2][1];
        currentData[i++] = basis.elements[2][2];
        currentData[i++] = point.z;
    }

    meshSphere->set_instance_count(static_cast<int64_t>(sphereCount));
    meshSphere->set_as_bulk_array(sphereData);
}

void
avi::scene::PathElementNode::setPosition(godot::Vector3 position)
{
    staticBody->set_translation(position);
}

void
avi::scene::PathElementNode::setRotation(godot::Quat rotation)
{
    godot::Transform quaternionTransform = godot::Transform(rotation);
    quaternionTransform.origin = staticBody->get_transform().origin;

    staticBody->set_transform(quaternionTransform);
}

void
avi::scene::PathElementNode::setScale(godot::Vector3 scale)
{
    staticBody->set_scale(scale);
}

void
avi::scene::PathElementNode::setColor(godot::Color color)
{
    material->set_albedo(color);

    if (color.a < 1)
    {
        material->set_feature(godot::SpatialMaterial::FEATURE_TRANSPARENT, true);
    }
}
