#include "PlayfulControlScheme.h"

#include <utility>

#include <Godot.hpp>
#include <Input.hpp>
#include <InputEvent.hpp>
#include <GlobalConstants.hpp>
#include <InputEventMouseMotion.hpp>
#include <InputEventMouseButton.hpp>
#include <Control.hpp>
#include <SceneTree.hpp>

void
avi::scene::PlayfulControlScheme::_register_methods()
{
    register_method("_ready", &PlayfulControlScheme::_ready);
    register_method("_unhandled_input", &PlayfulControlScheme::_unhandled_input);
    register_method("_process", &PlayfulControlScheme::_process);
}

void
avi::scene::PlayfulControlScheme::_init()
{
    viewCamera = nullptr;
    scrollValue = 0.0f;

    inputState = std::vector<bool>(MAX_INPUT_INDEX, false);

    inputKeys.push_back("avi_camera_forward");
    inputKeys.push_back("avi_camera_backward");
    inputKeys.push_back("avi_camera_left");
    inputKeys.push_back("avi_camera_right");
    inputKeys.push_back("avi_camera_up");
    inputKeys.push_back("avi_camera_down");
    inputKeys.push_back("avi_camera_roll_left");
    inputKeys.push_back("avi_camera_roll_right");
}

void
avi::scene::PlayfulControlScheme::setup(avi::core::SettingsProvider& settings,
                                        avi::scene::VisualizedScene& scene,
                                        bool gravity)
{
    this->settings = &settings;
    this->scene = &scene;
    this->gravity = gravity;
}

void
avi::scene::PlayfulControlScheme::_ready()
{
    attachment = get_node<godot::Spatial>("Attachment");
    head = get_node<godot::Spatial>("Attachment/Head");
}

void
avi::scene::PlayfulControlScheme::_unhandled_input(godot::InputEvent* event)
{
    if (!isActive()) return;

    auto* mouseMotionEvent = cast_to<godot::InputEventMouseMotion>(event);
    if (mouseMotionEvent)
    {
        mouseDelta = mouseMotionEvent->get_relative();
    }

    auto* mouseButtonEvent = cast_to<godot::InputEventMouseButton>(event);
    if (mouseButtonEvent)
    {
        float change = 0.0f;
        auto idx = mouseButtonEvent->get_button_index();

        if (idx == godot::GlobalConstants::BUTTON_WHEEL_UP)
        {
            change += 1.0f;
        }

        if (idx == godot::GlobalConstants::BUTTON_WHEEL_DOWN)
        {
            change -= 1.0f;
        }

        scrollValue += change;
        scrollValue = std::clamp(scrollValue, -100.0f, 200.0f);
    }

    for (auto i = 0; i < MAX_INPUT_INDEX; i++)
    {
        if (event->is_action_pressed(inputKeys[i]))
            inputState[i] = true;
    }
}

void
avi::scene::PlayfulControlScheme::_process(float delta)
{
    if (!isActive()) return;

    godot::Vector2 oldMouseDelta = mouseDelta;
    mouseDelta = {0, 0};

    if (isTraveling)
    {
        doTravel(delta);
        return;
    }

    godot::Input* input = godot::Input::get_singleton();

    for (auto i = 0; i < MAX_INPUT_INDEX; i++)
    {
        if (!input->is_action_pressed(inputKeys[i]))
            inputState[i] = false;
    }

    bool rotate = input->is_action_pressed("avi_camera_mouse") || viewCamera->isMouseLocked();

    float mouseSensitivity = static_cast<float>(settings->getMouseSensitivity()) * 0.001;

    if (rotate)
    {
        attachment->rotate_object_local({0, 1, 0},
                                        mouseSensitivity * oldMouseDelta.x * delta * -1.0f);

        if (gravity)
        {
            head->rotate_object_local({1, 0, 0},
                                      mouseSensitivity * oldMouseDelta.y * delta * -1.0f);
        }
        else
        {
            attachment->rotate_object_local({1, 0, 0},
                                            mouseSensitivity * oldMouseDelta.y * delta * -1.0f);
        }
    }

    if (rotatePressed && continuousEnabled)
    {
        attachment->rotate_object_local({0, 1, 0}, pressedRotation.x * delta * -3.0f);

        if (gravity)
        {
            head->rotate_object_local({1, 0, 0}, pressedRotation.y * delta * -3.0f);
        }
        else
        {
            attachment->rotate_object_local({1, 0, 0}, pressedRotation.y * delta * -3.0f);
        }
    }

    godot::Vector3 movement;

    if (viewCamera->isOrthogonal())
    {
        if (inputState[INPUT_FORWARD])
            movement += {0, 1, 0};
        if (inputState[INPUT_BACKWARD])
            movement += {0, -1, 0};
        if (inputState[INPUT_LEFT])
            movement += {-1, 0, 0};
        if (inputState[INPUT_RIGHT])
            movement += {1, 0, 0};
        if (inputState[INPUT_UP])
            viewCamera->set_size(
                (viewCamera->get_size() - delta < 0.001) ? 0.001 : viewCamera->get_size() - delta);
        if (inputState[INPUT_DOWN])
            viewCamera->set_size(viewCamera->get_size() + delta);
    }
    else
    {
        if (inputState[INPUT_FORWARD])
            movement += {0, 0, -1};
        if (inputState[INPUT_BACKWARD])
            movement += {0, 0, 1};
        if (inputState[INPUT_LEFT])
            movement += {-1, 0, 0};
        if (inputState[INPUT_RIGHT])
            movement += {1, 0, 0};
        if (inputState[INPUT_UP])
            movement += {0, 1, 0};
        if (inputState[INPUT_DOWN])
            movement += {0, -1, 0};
    }


    if (movement.length() > 0)
    {
        movement.normalize();
        movement *= delta * static_cast<float>(settings->getMoveSpeedPlayful() * 0.01);
        movement *= std::pow(1.05f, scrollValue);

        attachment->translate(movement);
    }

    if (!gravity)
    {
        float roll = 0;
        if (inputState[INPUT_ROLL_LEFT])
            roll += 1;
        if (inputState[INPUT_ROLL_RIGHT])
            roll -= 1;

        if (roll != 0)
        {
            roll *= delta * static_cast<float>(settings->getRollSensitivityPlayful() * 0.02);
            attachment->rotate_object_local({0, 0, 1}, roll);
        }
    }
}


void
avi::scene::PlayfulControlScheme::takeOver(avi::scene::ViewCamera* viewCamera,
                                           godot::Transform head)
{
    this->viewCamera = viewCamera;

    attachment->set_transform(viewCamera->get_transform());
    viewCamera->set_transform(godot::Transform::IDENTITY);
    this->head->set_transform(head);
    this->head->add_child(viewCamera);

    if (viewCamera->isOrthogonal())
    {
        advertiser->activateAdvertisements(adIDOrthographic);
        advertiser->deactivateAdvertisements(adID);
    }
    else
    {
        advertiser->activateAdvertisements(adID);
        advertiser->deactivateAdvertisements(adIDOrthographic);
    }
}

godot::Transform
avi::scene::PlayfulControlScheme::release()
{
    head->remove_child(viewCamera);
    viewCamera->set_transform(attachment->get_transform());
    this->viewCamera = nullptr;

    if (advertiser != nullptr)
    {
        advertiser->deactivateAdvertisements(adID);
        advertiser->deactivateAdvertisements(adIDOrthographic);
    }

    return head->get_transform();
}

void
avi::scene::PlayfulControlScheme::reset()
{
    attachment->set_transform(attachment->get_transform() * head->get_transform());
    head->set_transform(godot::Transform::IDENTITY);

    godot::Vector3 transformedX = scene->get_transform().basis.xform(godot::Vector3(1, 0, 0));
    startTravel(scene->get_transform().rotated(transformedX, Math_PI * 0.5), 0.5f);
}

void
avi::scene::PlayfulControlScheme::focus(godot::AABB box)
{
    attachment->set_transform(attachment->get_transform() * head->get_transform());
    head->set_transform(godot::Transform::IDENTITY);

    godot::Transform transform;
    godot::Vector3 aabbCenter = box.position + box.size * 0.5f;

    float zoomValue = box.size.length() * 3;
    godot::Vector3 zoomVector =
        (aabbCenter - attachment->get_global_transform().origin).normalized();

    transform.origin = aabbCenter - zoomVector * zoomValue;
    transform = transform.looking_at(box.position + box.size * 0.5f, {0, 1, 0});

    this->cameraStartSize = viewCamera->get_size();
    this->cameraEndSize = box.size.length();
    this->updateSize = true;

    startTravel(transform, 0.5f);

    float clipPlaneValue = zoomValue * 0.01f;
    this->viewCamera->setClipPlanes(clipPlaneValue);
}

void
avi::scene::PlayfulControlScheme::focus(godot::Vector3 point, avi::core::Element* target)
{
    attachment->set_transform(attachment->get_transform() * head->get_transform());
    head->set_transform(godot::Transform::IDENTITY);

    godot::AABB box = target->getVolume();
    godot::Transform transform;

    float zoomValue = box.size.length() * 3;
    godot::Vector3 zoomVector = (point - attachment->get_global_transform().origin).normalized();

    transform.origin = point - zoomVector * zoomValue;
    transform = transform.looking_at(box.position + box.size * 0.5f, {0, 1, 0});

    this->cameraStartSize = viewCamera->get_size();
    this->cameraEndSize = box.size.length() * 2;
    this->updateSize = true;

    startTravel(transform, 0.5f);

    float clipPlaneValue = zoomValue * 0.01f;
    this->viewCamera->setClipPlanes(clipPlaneValue);
}

godot::Basis
avi::scene::PlayfulControlScheme::getRotationBasis()
{
    godot::Transform transform = attachment->get_transform() * head->get_transform();

    return transform.basis;
}

bool
avi::scene::PlayfulControlScheme::isActive()
{
    return viewCamera != nullptr;
}

void
avi::scene::PlayfulControlScheme::startTravel(godot::Transform target, float duration)
{
    godot::Transform transformedTarget = target;
    godot::Transform transformHead;


    if (gravity)
    {
        float angleZ = transformedTarget.get_basis().get_euler().z;
        transformedTarget = transformedTarget.rotated(transformedTarget.get_basis().z, -angleZ);

        float angleX = transformedTarget.get_basis().get_euler().x;
        transformedTarget = transformedTarget.rotated(transformedTarget.get_basis().x, -angleX);
        transformHead.rotate({1, 0, 0}, angleX);

        transformedTarget.origin = target.origin;

        godot::Vector3 euler = transformedTarget.get_basis().get_euler();
        euler.z = 0;
        transformedTarget.basis.set_euler_xyz(euler);
    }

    travel = {attachment->get_transform() * head->get_transform(),
              std::move(transformedTarget),
              transformHead,
              duration,
              duration};
    isTraveling = true;
}

void
avi::scene::PlayfulControlScheme::doTravel(float delta)
{
    travel.remainingTime -= delta;
    float progress = 1.0f - (travel.remainingTime / travel.duration);

    if (travel.remainingTime < 0)
    {
        isTraveling = false;
        updateSize = false;
        progress = 1.0f;
    }

    godot::Transform transform =
        travel.startTransform.interpolate_with(travel.endTransform, progress);

    if (this->updateSize || !isTraveling)
    {
        viewCamera->set_size(
            godot::Math::lerp(this->cameraStartSize, this->cameraEndSize, progress));
    }

    godot::Transform transformHead =
        godot::Transform::IDENTITY.interpolate_with(travel.endTransformHead, progress);

    head->set_transform(transformHead);

    attachment->set_transform(transform);
}
void
avi::scene::PlayfulControlScheme::placeOnAxisNumpad(Axis axis, AxisSign sign)
{
    godot::Vector3 origin = attachment->get_transform().origin;
    godot::Vector3 transformedX = scene->get_transform().basis.xform(godot::Vector3(1, 0, 0));
    godot::Transform transform = scene->get_transform().rotated(transformedX, Math_PI * 0.5);

    if (axis == xAxis)
    {
        godot::Basis xBasis = godot::Basis(0, 0, -1, 0, 1, 0, 1, 0, 0);

        switch (sign)
        {
            case AxisSign::positive:
                transform.basis *= xBasis;
                break;
            case AxisSign::negative:
                transform.basis *= xBasis.inverse();
                break;
            case AxisSign::signSwitch:
                if (switchX)
                {
                    transform.basis *= xBasis.inverse();
                }
                else
                {
                    transform.basis *= xBasis;
                }
                switchX = !switchX;
                break;
        }
    }

    if (axis == yAxis)
    {
        godot::Basis yBasis = godot::Basis(-1, 0, 0, 0, 1, 0, 0, 0, -1);

        switch (sign)
        {
            case AxisSign::positive:
                transform.basis *= yBasis;
                break;
            case AxisSign::negative:
                break;
            case AxisSign::signSwitch:
                if (!switchY)
                {
                    transform.basis *= yBasis;
                }
                switchY = !switchY;
                break;
        }
    }

    if (axis == zAxis)
    {
        godot::Basis zBasis = godot::Basis(1, 0, 0, 0, 0, 1, 0, -1, 0);

        switch (sign)
        {
            case AxisSign::positive:
                transform.basis *= zBasis;
                break;
            case AxisSign::negative:
                transform.basis *= zBasis.inverse();
                break;
            case AxisSign::signSwitch:
                if (switchZ)
                {
                    transform.basis *= zBasis.inverse();
                }
                else
                {
                    transform.basis *= zBasis;
                }
                switchZ = !switchZ;
                break;
        }
    }

    transform.set_origin(origin);
    startTravel(transform, 0.3f);
}

void
avi::scene::PlayfulControlScheme::placeOnAxis(godot::Vector3 fromVector)
{
    attachment->set_transform(attachment->get_transform() * head->get_transform());
    head->set_transform(godot::Transform::IDENTITY);

    godot::Transform transform = attachment->get_transform() * head->get_transform();

    godot::Vector3 axis = fromVector.cross(godot::Vector3(0, 0, -1)).normalized();
    float angle = Math_PI - fromVector.angle_to(godot::Vector3(0, 0, -1));

    godot::Quat quaternion(axis.x * sin(angle * 0.5f),
                           axis.y * sin(angle * 0.5f),
                           axis.z * sin(angle * 0.5f),
                           cos(angle * 0.5f));

    godot::Transform rotation(quaternion);
    transform = transform * rotation;


    transform.origin = attachment->get_transform().origin;
    startTravel(transform, 0.3f);
}

void
avi::scene::PlayfulControlScheme::invertAxis(godot::Vector3 fromVector)
{
    godot::Transform transform = attachment->get_transform() * head->get_transform();

    float angleOffsetY =
        -0.5 * std::atan2((fromVector.cross(godot::Vector3(0, 0, -1))).dot(godot::Vector3(1, 0, 0)),
                          fromVector.dot(godot::Vector3(0, 0, -1)));
    godot::Vector3 axis = godot::Vector3(0, 1, 0).rotated(godot::Vector3(1, 0, 0), angleOffsetY);
    float signedAngle =
        std::atan2((fromVector.cross(godot::Vector3(0, 0, -1))).dot(godot::Vector3(0, 1, 0)),
                   fromVector.dot(godot::Vector3(0, 0, -1)));
    float angle = Math_PI - signedAngle;

    godot::Quat quaternion(axis.x * sin(angle * 0.5f),
                           axis.y * sin(angle * 0.5f),
                           axis.z * sin(angle * 0.5f),
                           cos(angle * 0.5f));
    godot::Transform rotation(quaternion);

    transform = transform * rotation;
    transform.origin = attachment->get_transform().origin;

    startTravel(transform, 0.3f);
}


void
avi::scene::PlayfulControlScheme::rotateFixed(RotationDirection direction, HotkeyRotation hotkey)
{
    double rotation = Math_PI * 0.0833;

    if (hotkey == CameraControlScheme::disable)
    {
        continuousEnabled = true;
        rotatePressed = false;
    }

    if (hotkey == CameraControlScheme::continuous)
    {
        rotatePressed = true;

        if (direction == NEGATIVE_HORIZONTAL || direction == NEGATIVE_VERTICAL)
        {
            rotation = -rotation;
        }

        if (direction == NEGATIVE_HORIZONTAL || direction == POSITIVE_HORIZONTAL)
        {
            pressedRotation = godot::Vector2(rotation, 0);
        }
        else
        {
            pressedRotation = godot::Vector2(0, rotation);
        }
    }

    if (hotkey == CameraControlScheme::fixed)
    {
        continuousEnabled = false;

        if (direction == NEGATIVE_HORIZONTAL || direction == NEGATIVE_VERTICAL)
        {
            rotation = -rotation;
        }

        if (direction == NEGATIVE_HORIZONTAL || direction == POSITIVE_HORIZONTAL)
        {
            attachment->rotate_object_local({0, 1, 0}, -rotation);
        }
        else
        {
            if (gravity)
            {
                head->rotate_object_local({1, 0, 0}, rotation);
            }
            else
            {
                attachment->rotate_object_local({1, 0, 0}, rotation);
            }
        }
    }
}

void
avi::scene::PlayfulControlScheme::setupControlAdvertisements(
    avi::core::ControlAdvertiser& advertiser)
{
    this->advertiser = &advertiser;

    using Ad = avi::core::ControlAdvertiser::Advertisement;

    {
        std::vector<avi::core::ControlAdvertiser::Advertisement> ads;
        ads.emplace_back(Ad::fromString("Playful Camera"));
        ads.emplace_back(Ad::fromString("WASD", "Move"));
        if (!gravity)
        {
            ads.emplace_back(Ad::fromString("QE", "Roll"));
        }
        ads.emplace_back(Ad::fromString("Space", "Up"));
        ads.emplace_back(Ad::fromString("X", "Down"));
        ads.emplace_back(Ad::fromString("Mouse Wheel Scroll", "Change Speed"));

        this->adID = advertiser.registerAdvertisement(std::move(ads));
    }

    {
        std::vector<avi::core::ControlAdvertiser::Advertisement> adsOrthographic;
        adsOrthographic.emplace_back(Ad::fromString("Playful Camera"));
        adsOrthographic.emplace_back(Ad::fromString("WASD", "Move"));
        if (!gravity)
        {
            adsOrthographic.emplace_back(Ad::fromString("QE", "Roll"));
        }
        adsOrthographic.emplace_back(Ad::fromString("Space", "Zoom in"));
        adsOrthographic.emplace_back(Ad::fromString("X", "Zoom out"));
        adsOrthographic.emplace_back(Ad::fromString("Mouse Wheel Scroll", "Change Speed"));

        this->adIDOrthographic = advertiser.registerAdvertisement(std::move(adsOrthographic));
    }

    if (this->isActive())
    {
        advertiser.activateAdvertisements(adID);
    }
}


void
avi::scene::PlayfulControlScheme::updateControlAdvertisements()
{
    if (viewCamera->isOrthogonal())
    {
        advertiser->activateAdvertisements(adIDOrthographic);
        advertiser->deactivateAdvertisements(adID);
    }
    else
    {
        advertiser->activateAdvertisements(adID);
        advertiser->deactivateAdvertisements(adIDOrthographic);
    }
}
